package org.apache.xindice.server;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: XindiceServlet.java,v 1.10 2002/12/13 09:02:19 gianugo Exp $
 */

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.Database;
import org.apache.xindice.server.rpc.RPCMessageInterface;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.xml.TextWriter;
import org.apache.xindice.xml.dom.DOMParser;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xmlrpc.XmlRpc;
import org.apache.xmlrpc.XmlRpcServer;

import org.w3c.dom.*;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.StringTokenizer;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * A <code>HttpServlet</code> that enables XML-RPC access to a Xindice
 * database instance.
 *
 * @author <a href="mailto:kstaken@xmldatabases.org">Kimbro Staken</a>
 * @author <a href="mailto:vladimir@apache.org">Vladimir R. Bossicard</a>
 * @author <a href="mailto:gianugo@apache.org">Gianugo Rabellino</a>
 */
public class XindiceServlet
      extends HttpServlet {

   protected static Log log = LogFactory.getLog("org.apache.xindice.servlet");

   protected XmlRpcServer xmlrpc;
   protected Database db;

   /**
    * TODO: verify that if is an error occured, the database will be closed
    * propertly
    */
   public void init(ServletConfig config)
         throws ServletException {

      /* note: there no need to call the DatabaseManager since we already
       * know the database we are using: the XML-RPC one */
      db = new Database();

      /* holds the database configuration */
      Document configuration;
      try {
         configuration = loadConfiguration(config);
      } catch (Exception e) {
         throw new ServletException("Could not load database configuration", e);
      }

      // The configuration is wrapped in a <xindice> element so we need to get
      // the "root-collection" configuration.
      Configuration conf = new Configuration(configuration, false);
      try {
         conf = conf.getChild("root-collection", false);
         if (conf != null) {
            
            // We need to ensure that the database points to a place where it makes
            // sense. If the path in the system.xml file is an absolute path, then
            // honor it. If it's not, we first check for the system property "xindice.db.home"
            // and if the lookup is successful we use it as the database root parent. If
            // the property is not set, we use /WEB-INF relative to the servlet context, unless
            // the war has not been unpacked. In this case, we throw an exception and 
            // ask the user to specify the location of database root
            
            String dbRoot = conf.getAttribute(Database.DBROOT, "./db/");
            File dbRootDir = new File(dbRoot);
            
            if (dbRootDir.isAbsolute()) {
            	db.setConfig(conf);	
            } else {            	
				// OK, no absolute path. We have to perform some checks.
				
                // Stupid hack but spec compliant. If getRealPath() returns null
                // the war archive has not been unpacked.
                String realPath = config.getServletContext().getRealPath("/WEB-INF");
				
				// Let's see if the property was specified.
				String home = System.getProperty("xindice.db.home");	            
	            
	            if (home != null) {	            	
	            	conf.setAttribute(Database.DBROOT, home + System.getProperty("file.separator") + dbRoot);
	            } else if (realPath != null) {
	            	if (log.isWarnEnabled()) 
	            		log.warn("The database root directory has been set to " 
	            			+ realPath + System.getProperty("file.separator") + dbRoot 
	            			+ ". Keep in mind that if a war upgrade will take place the database" 
	            			+ " will be lost.");
	            	conf.setAttribute(Database.DBROOT, realPath + System.getProperty("file.separator") + dbRoot);	
	            } else {	 	
	            	log.fatal("The database configuration points to a relative path, "
	            		+ "but there was no xindice.home property set. " 
	            		+ "Furthermore, the war was not unpacked by the application server "
	            		+ "so Xindice was unable to find a database location "
	            		+ "Please check /WEB-INF/system.xml and set an absolute path "
	            		+ "as the \"dbroot\" attribute of \"root-collection\" " 
	            		+ "or specify a suitable xindice.db.home system property.");
					throw new ServletException("An error occurred initializing the database." +
						"Please see the logs for details");	            	
	            }		            	
	            	 
            }  
            
                       
            db.setConfig(conf);
            
         } else {
            throw new ServletException("The database configuration is missing the <root-collection> element");
         }
      } catch (Exception e) {
         throw new ServletException("Error while handling the configuration", e);
      }

      // Setup the XML-RPC impl to support UTF-8 input via Xerces.
      XmlRpc.setEncoding("UTF8");
      try {
         // TODO: This introduces a firm dependency on xerces. fix it.
         XmlRpc.setDriver("xerces");
      } catch (Exception e) {
         throw new ServletException("XML-RPC support requires Xerces", e);
      }

      // Create the XML-RPC server and add our handler as the default.
      xmlrpc = new XmlRpcServer();
      try {
         xmlrpc.addHandler("$default", new RPCMessageInterface());
      } catch (Exception e) {
         throw new ServletException("Startup error", e);
      }

      if (log.isInfoEnabled()) {
         log.info("Database successfully started");
         //log.info("database location: '" + first.getFirstChild().getNodeValue() + "'");
      }
   }

   public void destroy() {
      try {
         // When the servlet engine goes down we need to close the
         // database instance.
         // TODO: make sure we're handling threads properly in this case.
         if (db != null) {
            db.close();
         }
         if (log.isInfoEnabled()) {
            log.info("Database successfully closed");
         }
      } catch (Exception e) {
         log.error("Error in destroy", e);
      }
   }

   /**
    * Loads the Xindice configuration file. The file is searched in the
    * following locations:
    * <ul>
    * <li>the <tt>xindice-configuration</tt> variable located in the servlet
    * configuration file</li>
    * <li>use the default configuration stored in the <tt>Xindice</tt> class</li>
    * </ul>
    * TODO: we should probably try to load from the file system if we
    * can't load it this way.
    */
   public Document loadConfiguration(ServletConfig config)
         throws XindiceException, IOException {
      Document result;

      String path = config.getInitParameter("xindice-configuration");
      if (path != null) {
         if (log.isDebugEnabled()) {
            log.debug("loading configuration from " + path);
         }
         ServletContext context = config.getServletContext();
         InputStream fis = context.getResourceAsStream(path);
         result = DOMParser.toDocument(fis);
         fis.close();
      } else {
         if (log.isDebugEnabled()) {
            log.debug("loading the standard configuration");
         }
         result = DOMParser.toDocument(Xindice.DEFAULT_CONFIGURATION);
      }

      return result;
   }

   // Get handling to provide simple HTTP access to the Xindice instance.
   public void doGet(HttpServletRequest request,
                     HttpServletResponse response)
         throws ServletException, IOException {
      try {
	 String contextPath = request.getContextPath();
         StringBuffer sb = new StringBuffer();
         String path = request.getQueryString();
         sb.append("<html><body><center>");
         if (path == null) {
            sb.append("<h2>THIS IS AN UGLY DEBUG TOOL!</h2><p>");
            sb.append("To browse the database, follow the link ");
            sb.append("<a href=\"" + contextPath + "?/" +
	       db.getName() + "\">" + db.getName() + "</a>");
         } else {
            // we have something to chew on
            XPathPseudoParser parser = new XPathPseudoParser(path);

            /* building the page once we have all the information */
            sb.append("<table border=\"1\" width=\"90%\">");
            sb.append("<tr><td rowspan=\"1\" colspan=\"2\">" + getPathNavigation(parser, contextPath) + "</td></tr>");
            sb.append("<tr><td valign=\"top\">" + getHierarchy(parser, contextPath) + "</td>");
            sb.append("<td valign=\"top\">" + getDetailView(parser) + "</td></tr>");
            sb.append("</table>");
         }
         sb.append("</center></body></html>");

         response.setContentType("text/html; charset=utf-8");
         byte[] resultBytes = sb.toString().getBytes("utf-8");
         response.setContentLength(resultBytes.length);
         OutputStream output = response.getOutputStream();
         output.write(resultBytes);
         output.flush();
      } catch (Exception e) {
         throw new ServletException("argh!", e);
         //response.setStatus(HttpServletResponse.SC_NOT_FOUND);
         // TODO: need intelligent error handling.
         // throw new ServletException("Error while doGet method", e);
      }
   }

   protected String getPathNavigation(XPathPseudoParser parser,
                                      String contextPath) {
      String path = parser.getQuery();
      StringBuffer result = new StringBuffer();
      StringTokenizer st = new StringTokenizer(path, "/");
      String currentPath = "<a href=\"" + contextPath + "?";
      while (st.hasMoreTokens()) {
         String token = st.nextToken();
         if (token.indexOf("&") > 0) {
            token = token.substring(0, token.indexOf("&"));
         }
         currentPath = currentPath + "/" + token;
         result.append(currentPath + "\">" + token + "</a>&nbsp;-&nbsp;");

      }
      return result.toString();
   }

   protected String getHierarchy(XPathPseudoParser parser,
                                 String contextPath)
         throws DBException, Exception {
      StringBuffer result = new StringBuffer();

      String path = parser.getPath();

      Collection col;
      if (!path.equals("")) {
         col = db.getCollection(parser.getPath());
      } else {
         col = db;
      }

      if (col == null) {
         result.append("Collection not found! " + parser.getPath());
         return result.toString();
      }

      String[] cols = col.listCollections();
      for (int i = 0; i < cols.length; i++) {
         result.append("<a href=\"" + contextPath + "?" +
            parser.getDatabase() + "/" + parser.getPath());
         result.append("/");
         result.append(cols[i]);
         result.append("\">");
         result.append(cols[i]);
         result.append("</a><br>");
      }

      try {
         String[] docs = col.listDocuments();
         for (int i = 0; i < docs.length; i++) {
            result.append("<a href=\"" + contextPath + "?" +
               parser.getDatabase() + "/" + parser.getPath());
            result.append("&");
            result.append(docs[i]);
            result.append("\">");
            result.append(docs[i]);
            result.append("</a><br>");
         }
      } catch (DBException e) {
         // do nothing but this collection cannot store documents.
      }

      return result.toString();
   }

   protected String getDetailView(XPathPseudoParser parser)
         throws Exception {
      Collection col = this.db.getCollection(parser.getPath());

      String document = parser.getDocument();
      if (document == null) {
         return "";
      }

      Document doc = col.getDocument(document);
      if (doc == null) {
         return "Document '" + document + "' not found";
      }

      return "Document '" + document + "'<p>" + TextWriter.toString(doc);
   }

   /**
    * Sends an XML query to the server and writes the output back. Currenlty
    * only XML-RPC query is supported.
    */
   public void doPost(HttpServletRequest request,
                      HttpServletResponse response)
         throws ServletException, IOException {
      byte[] result = xmlrpc.execute(request.getInputStream());
      response.setContentType("text/xml");
      response.setContentLength(result.length);
      OutputStream output = response.getOutputStream();
      output.write(result);
      output.flush();
   }

   public static class XPathPseudoParser {

      private String query;

      public XPathPseudoParser(String query)
            throws Exception {
         this.query = query;
      }

      public String getQuery() {
         return this.query;
      }

      public String getDatabase() {
         int max = query.indexOf("/", 1);
         if (max > 1) {
            return query.substring(0, max);
         } else {
            return query;
         }
      }

      public String getPath() {
         int min = query.indexOf("/", 1);
         if (min <= 0) {
            return "";
         }

         int max = query.indexOf("&");
         if (max > 1) {
            return query.substring(min + 1, max);
         } else {
            return query.substring(min + 1);
         }
         /*
         int pos = query.indexOf("&");
         if (pos > 0) {
            return query.substring(0, pos);
         } else {
            return query;
         }
         */
      }

      public String getDocument() {
         int pos = query.indexOf("&");
         if (pos > 0) {
            return query.substring(pos + 1);
         } else {
            return null;
         }
      }

   }

}
